@@ -3,7 +3,6 @@
 from __future__ import unicode_literals
 from unittest import skipIf, skipUnless
 
-from django import VERSION as django_version
 from django.contrib.auth.models import User, Permission, Group
 from django.contrib.contenttypes.models import ContentType
 from django.core.exceptions import MultipleObjectsReturned
@@ -18,9 +17,6 @@ from .models import Test, TestParent, TestChild
 from .test_utils import TestUtilsMixin
 
 
-DJANGO_GTE_1_11 = django_version[:2] >= (1, 11)
-
-
 class WriteTestCase(TestUtilsMixin, TransactionTestCase):
     """
     Tests if every SQL request writing data is not cached and invalidates the
@@ -84,9 +80,7 @@ class WriteTestCase(TestUtilsMixin, TransactionTestCase):
         with self.assertNumQueries(1):
             self.assertListEqual(list(Test.objects.all()), [])
 
-        with self.assertNumQueries(
-                (5 if self.is_sqlite else 4) if DJANGO_GTE_1_11
-                else (3 if self.is_sqlite else 2)):
+        with self.assertNumQueries(5 if self.is_sqlite else 4):
             t, created = Test.objects.update_or_create(
                 name='test', defaults={'public': True})
             self.assertTrue(created)
@@ -1038,7 +1032,7 @@ class DatabaseCommandTestCase(TestUtilsMixin, TransactionTestCase):
             self.assertListEqual(list(Test.objects.all()), [self.t])
 
         call_command('loaddata', 'cachalot/tests/loaddata_fixture.json',
-                     verbosity=0, interactive=False)
+                     verbosity=0)
 
         self.force_repoen_connection()
 
